#!/usr/bin/env python
# coding: utf-8

# In[ ]:


get_ipython().system('pip install -q kaggle')


# In[ ]:


#1. K-MEANS CLUSTERING

# Import the required libraries
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import math

# Kaggle dataset setup
get_ipython().system('mkdir -p ~/.kaggle')
get_ipython().system('cp kaggle.json ~/.kaggle/')
get_ipython().system('chmod 600 ~/.kaggle/kaggle.json')

get_ipython().system('kaggle datasets download -d uciml/iris')
get_ipython().system('unzip iris.zip')


# Load the Iris dataset
iris = pd.read_csv('Iris.csv')

#  Select the feature columns (SepalLengthCm, SepalWidthCm)
X = iris.iloc[:, 1:3].values  # Only SepalLengthCm and SepalWidthCm

# Function to calculate Euclidean distance
def euclidean_distance(p1, p2):
    return np.sqrt(np.sum((p1 - p2) ** 2))

# Initialize centroids far from each other
np.random.seed(0)  # For reproducibility
initial_idx = np.random.choice(len(X))
centroid1 = X[initial_idx]

distances = [euclidean_distance(x, centroid1) for x in X]
second_idx = np.argmax(distances)
centroid2 = X[second_idx]

distances = [min(euclidean_distance(x, centroid1), euclidean_distance(x, centroid2)) for x in X]
third_idx = np.argmax(distances)
centroid3 = X[third_idx]

sepal_length1, sepal_width1 = centroid1
sepal_length2, sepal_width2 = centroid2
sepal_length3, sepal_width3 = centroid3

print("Initial Centroids:")
print(f"Centroid 1: {centroid1}")
print(f"Centroid 2: {centroid2}")
print(f"Centroid 3: {centroid3}")
print("-------------------------------------")

# K-Means Clustering Loop
cluster1, cluster2, cluster3 = [], [], []
prev1, prev2, prev3 = [], [], []
count=0

while True:
    cluster1.clear()
    cluster2.clear()
    cluster3.clear()

    # Assign points to the nearest cluster
    for i in range(len(X)):
        sepal_length, sepal_width = X[i]
        dist1 = math.sqrt((sepal_length1 - sepal_length) ** 2 + (sepal_width1 - sepal_width) ** 2)
        dist2 = math.sqrt((sepal_length2 - sepal_length) ** 2 + (sepal_width2 - sepal_width) ** 2)
        dist3 = math.sqrt((sepal_length3 - sepal_length) ** 2 + (sepal_width3 - sepal_width) ** 2)

        if dist1 < dist2 and dist1 < dist3:
            cluster1.append(i)
        elif dist2 < dist3:
            cluster2.append(i)
        else:
            cluster3.append(i)

    # Check for convergence
    if prev1 == cluster1 and prev2 == cluster2 and prev3 == cluster3:
        break

    # Update centroids
    prev1, prev2, prev3 = cluster1[:], cluster2[:], cluster3[:]
    sepal_length1 = np.mean([X[i][0] for i in cluster1])
    sepal_width1 = np.mean([X[i][1] for i in cluster1])
    sepal_length2 = np.mean([X[i][0] for i in cluster2])
    sepal_width2 = np.mean([X[i][1] for i in cluster2])
    sepal_length3 = np.mean([X[i][0] for i in cluster3])
    sepal_width3 = np.mean([X[i][1] for i in cluster3])
    count=count+1

    print(f"Updated Centroids:")
    print(f"Centroid 1: Length={sepal_length1:.2f}, Width={sepal_width1:.2f}")
    print(f"Centroid 2: Length={sepal_length2:.2f}, Width={sepal_width2:.2f}")
    print(f"Centroid 3: Length={sepal_length3:.2f}, Width={sepal_width3:.2f}")
    print("-------------------------------------")

print("The clustering converges after ",count," iterations")

# Visualization of Clusters
plt.figure(figsize=(10, 6))
plt.scatter([X[i][0] for i in cluster1], [X[i][1] for i in cluster1], c='red', label='Cluster 1')
plt.scatter([X[i][0] for i in cluster2], [X[i][1] for i in cluster2], c='blue', label='Cluster 2')
plt.scatter([X[i][0] for i in cluster3], [X[i][1] for i in cluster3], c='green', label='Cluster 3')

# Plot the centroids
plt.scatter(sepal_length1, sepal_width1, c='black', marker='x', s=200, label='Centroids')
plt.scatter(sepal_length2, sepal_width2, c='black', marker='x', s=200)
plt.scatter(sepal_length3, sepal_width3, c='black', marker='x', s=200)

plt.xlabel('Sepal Length (cm)')
plt.ylabel('Sepal Width (cm)')
plt.title('K-Means Clustering of Iris Dataset')
plt.legend()
plt.show()





# In[ ]:


# 2. Draw a decision tree using ID3 algorithm on playgolf dataset.

get_ipython().system('pip install -q kaggle')
from google.colab import files
files.upload()
get_ipython().system('mkdir ~/.kaggle')
get_ipython().system('cp kaggle.json ~/.kaggle/')
get_ipython().system('chmod 600 ~/.kaggle/kaggle.json')
get_ipython().system('kaggle datasets download mmellinger66/playgolf-naivebayes')
get_ipython().system('unzip playgolf-naivebayes.zip')


import pandas as pd
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.tree import export_graphviz
import graphviz


# Load the dataset
data = pd.read_csv('PlayGolf.csv')

# Convert categorical features to numerical using one-hot encoding
data = pd.get_dummies(data, columns=['Outlook', 'Temperature', 'Humidity', 'Windy'])

# Separate features (X) and target (y)
X = data.drop('PlayGolf', axis=1)
y = data['PlayGolf']

# Split the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

# Create and train the Decision Tree classifier
clf = DecisionTreeClassifier(criterion='entropy')  # Use ID3 (entropy)
clf.fit(X_train, y_train)

# Make predictions on the test set
y_pred = clf.predict(X_test)

# Evaluate the model
accuracy = accuracy_score(y_test, y_pred)
print(f"Accuracy: {accuracy}")

# Visualize the decision tree
dot_data = export_graphviz(clf, out_file=None,
                           feature_names=X.columns,
                           class_names=['No', 'Yes'],
                           filled=True, rounded=True, special_characters=True)
graph = graphviz.Source(dot_data)
graph.render("playgolf_decision_tree")  # Save the tree as a PDF file
graph


# In[ ]:


# 3.Apply CART algorithm and draw the decision on buy computer dataset.


from google.colab import files
import pandas as pd
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.tree import export_graphviz
import graphviz
get_ipython().system('pip install -q kaggle')
files.upload()
get_ipython().system(' mkdir ~/.kaggle')
get_ipython().system(' cp kaggle.json ~/.kaggle/')
get_ipython().system(' chmod 600 ~/.kaggle/kaggle.json')
get_ipython().system(' kaggle datasets download usman27/buy-computer')
get_ipython().system(' unzip buy-computer.zip')


# Load the dataset (assuming the CSV file is named 'buy_computer.csv')
data = pd.read_csv('Buy_Computer.csv')

# Assuming 'buys_computer' is the target variable
# Convert categorical features to numerical using one-hot encoding
data = pd.get_dummies(data, columns=['age', 'income', 'student', 'credit_rating'])

# Separate features (X) and target (y)
X = data.drop('Buy_Computer', axis=1)
y = data['Buy_Computer']

# Split the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

# Create and train the Decision Tree classifier
clf = DecisionTreeClassifier(criterion='gini')  # Use CART (gini)
clf.fit(X_train, y_train)

# Make predictions on the test set
y_pred = clf.predict(X_test)

# Evaluate the model
accuracy = accuracy_score(y_test, y_pred)
print(f"Accuracy: {accuracy}")

# Visualize the decision tree
dot_data = export_graphviz(clf, out_file=None,
                           feature_names=X.columns,
                           class_names=['No', 'Yes'],  # Adjust class names as needed
                           filled=True, rounded=True, special_characters=True)
graph = graphviz.Source(dot_data)
graph.render("buy_computer_decision_tree")  # Save the tree as a PDF file
graph


# In[ ]:


# 4. Naive bayes classification on buys computer dataset.

from google.colab import files
from sklearn.naive_bayes import GaussianNB
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, classification_report
from sklearn.preprocessing import LabelEncoder, MinMaxScaler
import pandas as pd
get_ipython().system('pip install -q kaggle')
files.upload()
get_ipython().system(' mkdir ~/.kaggle')
get_ipython().system(' cp kaggle.json ~/.kaggle/')
get_ipython().system(' chmod 600 ~/.kaggle/kaggle.json')
get_ipython().system(' kaggle datasets download usman27/buy-computer')
get_ipython().system(' unzip buy-computer.zip')


# Load the dataset
data = pd.read_csv('Buy_Computer.csv')
print(data)

# Encode the target variable (Buy_Computer) if it is categorical
label_encoder = LabelEncoder()
data['Buy_Computer'] = label_encoder.fit_transform(data['Buy_Computer'])

# Convert categorical features to numerical using one-hot encoding
data = pd.get_dummies(data, columns=['age', 'income', 'student', 'credit_rating'], drop_first=True)

# Separate features (X) and target (y)
X = data.drop('Buy_Computer', axis=1)
y = data['Buy_Computer']

# Scale features if necessary
scaler = MinMaxScaler()
X = scaler.fit_transform(X)

# Split the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42, stratify=y)

# Create and train the Naive Bayes classifier
model = GaussianNB()
model.fit(X_train, y_train)

# Make predictions on the test set
y_pred = model.predict(X_test)

# Evaluate the model
accuracy = accuracy_score(y_test, y_pred)
print(f"Accuracy: {accuracy:.2f}")


# Function to get user input and predict
def get_user_input():
    age = input("Enter age (youth/middle_aged/senior): ")
    income = input("Enter income (high/medium/low): ")
    student = input("Enter student (yes/no): ")
    credit_rating = input("Enter credit rating (fair/excellent): ")

    # Create a dictionary to store the input values
    user_data = {
        'age_' + age: 1,
        'income_' + income: 1,
        'student_' + student: 1,
        'credit_rating_' + credit_rating: 1
    }

    # Create a DataFrame from the user input
    user_df = pd.DataFrame([user_data])

    # Ensure column order matches the training data
    user_df = user_df.reindex(columns=data.columns.drop('Buy_Computer'), fill_value=0)

    # Scale user input data
    user_df = scaler.transform(user_df)

    return user_df

#  user input for an unknown sample
user_input = get_user_input()

# Predict the class label for the unknown sample
predicted_class = model.predict(user_input)
predicted_label = label_encoder.inverse_transform(predicted_class)
print(f"Predicted class for the unknown sample - Buy's Computer: {predicted_label[0]}")


# In[ ]:


# 5. FUZZY C MEANS

from google.colab import files
import pandas as pd
from sklearn.preprocessing import MinMaxScaler
from fcmeans import FCM
import matplotlib.pyplot as plt
import seaborn as sns


get_ipython().system('pip install -q fuzzy-c-means')
get_ipython().system('pip install -q kaggle')
files.upload()
get_ipython().system('mkdir ~/.kaggle')
get_ipython().system('cp kaggle.json ~/.kaggle/')
get_ipython().system('chmod 600 ~/.kaggle/kaggle.json')

get_ipython().system('kaggle datasets download harlfoxem/housesalesprediction')
get_ipython().system('unzip -q housesalesprediction.zip')

# Load the dataset
data = pd.read_csv('kc_house_data.csv')

# Select relevant features for clustering
features = ['bedrooms', 'bathrooms', 'sqft_living', 'sqft_lot', 'floors', 'price']
X = data[features]

# Normalize the data using MinMaxScaler
scaler = MinMaxScaler()
X_scaled = scaler.fit_transform(X)

# Apply Fuzzy C-Means clustering
n_clusters = 3  # Number of clusters
fcm = FCM(n_clusters=n_clusters)
fcm.fit(X_scaled)

# Get cluster labels and membership values
labels = fcm.predict(X_scaled)
membership = fcm.u

# Add cluster labels to the original dataset
data['cluster'] = labels

# Define colors for each cluster
cluster_colors = sns.color_palette("Set1", n_colors=n_clusters)

# Descriptive statistics for each cluster
descriptive_stats = {}
for cluster_id in range(n_clusters):
    cluster_data = data[data['cluster'] == cluster_id]
    descriptive_stats[cluster_id] = {
        'sqft_living_mean': cluster_data['sqft_living'].mean(),
        'sqft_living_median': cluster_data['sqft_living'].median(),
        'sqft_living_std': cluster_data['sqft_living'].std(),
        'price_mean': cluster_data['price'].mean(),
        'price_median': cluster_data['price'].median(),
        'price_std': cluster_data['price'].std(),
    }
    print(f"Cluster {cluster_id + 1} Descriptive Statistics:")
    print(f"  Mean Sqft Living: {descriptive_stats[cluster_id]['sqft_living_mean']:.2f}")
    print(f"  Median Sqft Living: {descriptive_stats[cluster_id]['sqft_living_median']:.2f}")
    print(f"  Std Dev Sqft Living: {descriptive_stats[cluster_id]['sqft_living_std']:.2f}")
    print(f"  Mean Price: ${descriptive_stats[cluster_id]['price_mean']:.2f}")
    print(f"  Median Price: ${descriptive_stats[cluster_id]['price_median']:.2f}")
    print(f"  Std Dev Price: ${descriptive_stats[cluster_id]['price_std']:.2f}")
    print()

# Plot the clusters based on two features with enhanced visualization
plt.figure(figsize=(12, 8))
for cluster_id in range(n_clusters):
    cluster_points = data[data['cluster'] == cluster_id]
    plt.scatter(cluster_points['sqft_living'], cluster_points['price'],
                label=f'Cluster {cluster_id + 1}',
                color=cluster_colors[cluster_id], alpha=0.7, s=50)


# Add axis labels, title, legend, and grid
plt.xlabel('Square Feet Living Area', fontsize=14)
plt.ylabel('Price (in USD)', fontsize=14)
plt.title('Fuzzy C-Means Clustering of Housing Dataset', fontsize=16)
plt.legend(title="Clusters", fontsize=12, title_fontsize=14)
plt.grid(color='gray', linestyle='--', linewidth=0.5)
plt.show()





# In[ ]:


# 6.  Apply perceptron learning for realization of logic gates. (bias = 1)

import numpy as np
from sklearn.model_selection import train_test_split

# Activation function: Bipolar Step Function
def bipolar_step(x):
    return 1 if x >= 0 else -1

# Perceptron class
class Perceptron:
    def __init__(self, input_size, learning_rate=0.1, epochs=100):
        self.weights = np.zeros(input_size + 1)  # Including bias weight
        self.learning_rate = learning_rate
        self.epochs = epochs

    def train(self, X, Y):
        for epoch in range(self.epochs):
            for inputs, label in zip(X, Y):
                net_input = np.dot(inputs, self.weights[1:]) + self.weights[0]  # w.x + bias
                prediction = bipolar_step(net_input)
                error = label - prediction

                # Update weights and bias
                self.weights[1:] += self.learning_rate * error * inputs
                self.weights[0] += self.learning_rate * error

    def predict(self, inputs):
        net_input = np.dot(inputs, self.weights[1:]) + self.weights[0]
        return bipolar_step(net_input)

# Training data for bipolar AND, OR, NAND, NOR
bipolar_inputs = np.array([
    [-1, -1],
    [-1,  1],
    [ 1, -1],
    [ 1,  1]
])

# Target outputs for AND, OR, NAND, and NOR gates
bipolar_AND_outputs = np.array([-1, -1, -1, 1])
bipolar_OR_outputs = np.array([-1, 1, 1, 1])
bipolar_NAND_outputs = np.array([1, 1, 1, -1])
bipolar_NOR_outputs = np.array([1, -1, -1, -1])

# Function to train and test perceptron
def train_and_test_perceptron(inputs, outputs, gate_name):
    X_train, X_test, y_train, y_test = train_test_split(inputs, outputs, test_size=0.2, random_state=42)
    perceptron = Perceptron(input_size=2)
    perceptron.train(X_train, y_train)

    print(f"\n{gate_name} Gate:")
    print("Training Results:")
    for x, label in zip(X_train, y_train):
        print(f"Input: {x}, Target: {label}, Prediction: {perceptron.predict(x)}")

    print("\nTesting Results:")
    for x, label in zip(X_test, y_test):
        print(f"Input: {x}, Target: {label}, Prediction: {perceptron.predict(x)}")

# Train and test Perceptron for each gate
train_and_test_perceptron(bipolar_inputs, bipolar_AND_outputs, "Bipolar AND")
train_and_test_perceptron(bipolar_inputs, bipolar_OR_outputs, "Bipolar OR")
train_and_test_perceptron(bipolar_inputs, bipolar_NAND_outputs, "Bipolar NAND")
train_and_test_perceptron(bipolar_inputs, bipolar_NOR_outputs, "Bipolar NOR")


# In[ ]:


# 7. Apply back propagation algorithm on sample {1,0,1} with the class label {1,0}.
# (Where network topology: 3-2-2-2, all biases and weights are initialized at 0).
import numpy as np

# Activation function: Sigmoid
def sigmoid(x):
    return 1 / (1 + np.exp(-x))

# Derivative of sigmoid
def sigmoid_derivative(x):
    return x * (1 - x)

# Initialize the neural network
def initialize_network(input_size, hidden_layers, output_size):
    np.random.seed(0)
    network = []

    # Initialize weights and biases for each layer
    layer_sizes = [input_size] + hidden_layers + [output_size]
    for i in range(len(layer_sizes) - 1):
        weights = np.zeros((layer_sizes[i], layer_sizes[i + 1]))  # All weights set to 0
        biases = np.zeros((1, layer_sizes[i + 1]))  # All biases set to 0
        network.append({"weights": weights, "biases": biases})

    return network

# Forward pass
def forward_pass(network, inputs):
    activations = [inputs]
    for layer in network:
        net_input = np.dot(activations[-1], layer["weights"]) + layer["biases"]
        activation = sigmoid(net_input)
        activations.append(activation)
    return activations

# Backward pass
def backward_pass(network, activations, target):
    errors = []
    deltas = []

    # Calculate output error
    error = target - activations[-1]
    errors.append(error)
    delta = error * sigmoid_derivative(activations[-1])
    deltas.append(delta)

    # Backpropagate the error
    for i in reversed(range(len(network) - 1)):
        error = np.dot(deltas[-1], network[i + 1]["weights"].T)
        delta = error * sigmoid_derivative(activations[i + 1])
        errors.append(error)
        deltas.append(delta)

    deltas.reverse()
    return deltas

# Update weights and biases
def update_weights(network, activations, deltas, learning_rate):
    for i in range(len(network)):
        layer = network[i]
        layer["weights"] += learning_rate * np.dot(activations[i].T, deltas[i])
        layer["biases"] += learning_rate * np.sum(deltas[i], axis=0, keepdims=True)

# Train the neural network
def train_network(network, inputs, target, epochs, learning_rate):
    for epoch in range(epochs):
        activations = forward_pass(network, inputs)
        deltas = backward_pass(network, activations, target)
        update_weights(network, activations, deltas, learning_rate)
        if epoch % 100 == 0 or epoch == epochs - 1:
            loss = np.mean((target - activations[-1]) ** 2)
            print(f"Epoch {epoch + 1}/{epochs}, Loss: {loss:.6f}")

# Input data and target
inputs = np.array([[1, 0, 1]])  # Sample {1, 0, 1}
target = np.array([[1, 0]])    # Class label {1, 0}

# Network topology
input_size = 3
hidden_layers = [2, 2]
output_size = 2

# Initialize the network
network = initialize_network(input_size, hidden_layers, output_size)

# Train the network
train_network(network, inputs, target, epochs=1000, learning_rate=0.1)

# Final prediction
def predict(network, inputs):
    activations = forward_pass(network, inputs)
    return activations[-1]

output = predict(network, inputs)
print("Final Output:", output)
print("Actual Target:", target)


# In[ ]:


# 8. MADALINE FOR BIPOLAR XOR
import numpy as np
from sklearn.model_selection import train_test_split

# XOR Input and Target
data = np.array([
    [-1, -1],  # Input 1
    [-1, 1],   # Input 2
    [1, -1],   # Input 3
    [1, 1]     # Input 4
])
targets = np.array([-1, 1, 1, -1])  # XOR outputs in bipolar format

# Initialize weights
v0, v1, v2 = 0.5, 0.5, 0.5  # Weights of the output layer
w = np.random.uniform(-0.1, 0.1, (2, 3))  # Weights of the hidden layer (2 hidden units, 2 inputs + 1 bias)
bias = 1
learning_rate = 0.5

# Split the data into training and testing sets (80-20 split)
train_data, test_data, train_targets, test_targets = train_test_split(data, targets, test_size=0.2, random_state=42)

# Activation function (bipolar step function)
def bipolar_step(x):
    return 1 if x >= 0 else -1

# MADALINE Algorithm Implementation
def train_madaline(data, targets, v0, v1, v2, w, bias, learning_rate, max_epochs=100):
    epoch = 0
    while epoch < max_epochs:
        weight_updated = False
        for i, x in enumerate(data):

            x1, x2 = x
            zin1 = bias * w[0][0] + x1 * w[0][1] + x2 * w[0][2]
            zin2 = bias * w[1][0] + x1 * w[1][1] + x2 * w[1][2]
            zout1 = bipolar_step(zin1)
            zout2 = bipolar_step(zin2)

            yin = bias * v0 + zout1 * v1 + zout2 * v2
            yout = bipolar_step(yin)

            # Print training output
            print(f"Training - Input: {x}, Predicted: {yout}, Target: {targets[i]}")

            # Check and adjust weights if required
            t = targets[i]
            if yout != t:
                weight_updated = True
                if t == 1:  # Case I
                    if abs(zin1) < abs(zin2):
                        j = 0
                    else:
                        j = 1
                    w[j][0] += learning_rate * (1 - zin1 if j == 0 else    1 - zin2) * bias
                    w[j][1] += learning_rate * (1 - zin1 if j == 0 else 1 - zin2) * x1
                    w[j][2] += learning_rate * (1 - zin1 if j == 0 else 1 - zin2) * x2
                elif t == -1:  # Case II
                    if zout1 > 0:
                        w[0][0] += learning_rate * (1 - zin1) * bias
                        w[0][1] += learning_rate * (1 - zin1) * x1
                        w[0][2] += learning_rate * (1 - zin1) * x2
                    if zout2 > 0:
                        w[1][0] += learning_rate * (1 - zin2) * bias
                        w[1][1] += learning_rate * (1 - zin2) * x1
                        w[1][2] += learning_rate * (1 - zin2) * x2

        # Check stopping condition
        if not weight_updated:
            break
        epoch += 1

    return w, epoch

# Testing the MADALINE
def test_madaline(data, targets, v0, v1, v2, w, bias):
    correct = 0
    for i, x in enumerate(data):
        x1, x2 = x
        zin1 = bias * w[0][0] + x1 * w[0][1] + x2 * w[0][2]
        zin2 = bias * w[1][0] + x1 * w[1][1] + x2 * w[1][2]
        zout1 = bipolar_step(zin1)
        zout2 = bipolar_step(zin2)

        yin = bias * v0 + zout1 * v1 + zout2 * v2
        yout = bipolar_step(yin)

        # Print testing output
        print(f"Testing - Input: {x}, Predicted: {yout}, Target: {targets[i]}")

        if yout == targets[i]:
            correct += 1

    accuracy = (correct / len(data)) * 100
    return accuracy

# Train the network
w, epochs = train_madaline(train_data, train_targets, v0, v1, v2, w, bias, learning_rate)

# Test the network
accuracy = test_madaline(test_data, test_targets, v0, v1, v2, w, bias)

# Print results
print(f"Training completed in {epochs} epochs.")
print(f"Testing Accuracy: {accuracy:.2f}%")


# In[ ]:


# 9. AGNES
# Import necessary libraries
import numpy as np
import pandas as pd
from sklearn.datasets import load_iris
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering
import scipy.cluster.hierarchy as sch
import matplotlib.pyplot as plt

# Load the Iris dataset
iris = load_iris()
X = iris.data

# Scale the data (important for clustering)
scaler = StandardScaler()
X_scaled = scaler.fit_transform(X)

# Apply Agglomerative Clustering (AGNES) with different linkage methods
linkage_methods = ['single', 'complete', 'average']

# Plot dendrograms for each linkage method
fig, axes = plt.subplots(1, 3, figsize=(18, 6))

for i, linkage in enumerate(linkage_methods):
    # Apply AgglomerativeClustering (AGNES) model
    agnes = AgglomerativeClustering(linkage=linkage, distance_threshold=0, n_clusters=None) # Set distance_threshold=0, n_clusters=None for linkage matrix
    agnes.fit(X_scaled)

    # Get linkage matrix directly from the model
    linkage_matrix = sch.linkage(X_scaled, method=linkage)

    # Create a dendrogram
    sch.dendrogram(linkage_matrix, ax=axes[i], above_threshold_color='blue')
    axes[i].set_title(f'Dendrogram ({linkage.capitalize()} Linkage)')
    axes[i].set_xlabel('Sample Index')
    axes[i].set_ylabel('Distance')

# Show the plot
plt.tight_layout()
plt.show()


# In[ ]:


#10. K MEDIOIDS

# Import the required libraries
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import math

# Kaggle dataset setup
get_ipython().system('mkdir -p ~/.kaggle')
get_ipython().system('cp kaggle.json ~/.kaggle/')
get_ipython().system('chmod 600 ~/.kaggle/kaggle.json')

get_ipython().system('kaggle datasets download -d uciml/iris')
get_ipython().system('unzip iris.zip')

# Load the Iris dataset
iris = pd.read_csv('Iris.csv')

# Select the feature columns (SepalLengthCm, SepalWidthCm)
X = iris.iloc[:, 1:3].values  # Only SepalLengthCm and SepalWidthCm

# Function to calculate Euclidean distance
def euclidean_distance(p1, p2):
    return np.sqrt(np.sum((p1 - p2) ** 2))

# Initialize medoids far from each other
np.random.seed(0)  # For reproducibility
initial_idx = np.random.choice(len(X))
medoid1 = X[initial_idx]

distances = [euclidean_distance(x, medoid1) for x in X]
second_idx = np.argmax(distances)
medoid2 = X[second_idx]

distances = [min(euclidean_distance(x, medoid1), euclidean_distance(x, medoid2)) for x in X]
third_idx = np.argmax(distances)
medoid3 = X[third_idx]

sepal_length1, sepal_width1 = medoid1
sepal_length2, sepal_width2 = medoid2
sepal_length3, sepal_width3 = medoid3

print("Initial Medoids:")
print(f"Medoid 1: {medoid1}")
print(f"Medoid 2: {medoid2}")
print(f"Medoid 3: {medoid3}")
print("-------------------------------------")

# K-Medoids Clustering Loop
cluster1, cluster2, cluster3 = [], [], []
prev1, prev2, prev3 = [], [], []
count = 0

while True:
    cluster1.clear()
    cluster2.clear()
    cluster3.clear()

    # Assign points to the nearest medoid
    for i in range(len(X)):
        sepal_length, sepal_width = X[i]
        dist1 = euclidean_distance([sepal_length1, sepal_width1], X[i])
        dist2 = euclidean_distance([sepal_length2, sepal_width2], X[i])
        dist3 = euclidean_distance([sepal_length3, sepal_width3], X[i])

        if dist1 < dist2 and dist1 < dist3:
            cluster1.append(i)
        elif dist2 < dist3:
            cluster2.append(i)
        else:
            cluster3.append(i)

    # Check for convergence
    if prev1 == cluster1 and prev2 == cluster2 and prev3 == cluster3:
        break

    # Update medoids
    prev1, prev2, prev3 = cluster1[:], cluster2[:], cluster3[:]

    def find_new_medoid(cluster):
        best_medoid = cluster[0]
        best_cost = float('inf')
        for candidate_idx in cluster:
            cost = sum(euclidean_distance(X[candidate_idx], X[j]) for j in cluster)
            if cost < best_cost:
                best_cost = cost
                best_medoid = candidate_idx
        return X[best_medoid]

    medoid1 = find_new_medoid(cluster1)
    medoid2 = find_new_medoid(cluster2)
    medoid3 = find_new_medoid(cluster3)

    sepal_length1, sepal_width1 = medoid1
    sepal_length2, sepal_width2 = medoid2
    sepal_length3, sepal_width3 = medoid3

    count += 1

    print(f"Updated Medoids:")
    print(f"Medoid 1: Length={sepal_length1:.2f}, Width={sepal_width1:.2f}")
    print(f"Medoid 2: Length={sepal_length2:.2f}, Width={sepal_width2:.2f}")
    print(f"Medoid 3: Length={sepal_length3:.2f}, Width={sepal_width3:.2f}")
    print("-------------------------------------")

print("The clustering converges after ", count, " iterations")

# Visualization of Clusters
plt.figure(figsize=(10, 6))
plt.scatter([X[i][0] for i in cluster1], [X[i][1] for i in cluster1], c='red', label='Cluster 1')
plt.scatter([X[i][0] for i in cluster2], [X[i][1] for i in cluster2], c='blue', label='Cluster 2')
plt.scatter([X[i][0] for i in cluster3], [X[i][1] for i in cluster3], c='green', label='Cluster 3')

# Plot the medoids
plt.scatter(sepal_length1, sepal_width1, c='black', marker='x', s=200, label='Medoids')
plt.scatter(sepal_length2, sepal_width2, c='black', marker='x', s=200)
plt.scatter(sepal_length3, sepal_width3, c='black', marker='x', s=200)

plt.xlabel('Sepal Length (cm)')
plt.ylabel('Sepal Width (cm)')
plt.title('K-Medoids Clustering of Iris Dataset')
plt.legend()
plt.show()


# In[ ]:


# 11. MADALINE VARIABLE TOPOLOGY

import numpy as np

class MadelineNetwork:
    def __init__(self, input_size, hidden_size, learning_rate=0.1):
        # Initialize weights with small random values
        self.learning_rate = learning_rate
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weights_input_hidden = np.random.uniform(-0.1, 0.1, (hidden_size, input_size + 1))  # +1 for bias
        self.weights_hidden_output = np.random.uniform(-0.1, 0.1, hidden_size + 1)  # +1 for bias

    def bipolar_step(self, x):
        return 1 if x >= 0 else -1

    def forward_propagation(self, inputs):
        # Add bias input
        inputs = np.append(1, inputs)  # Bias at the front

        # Compute hidden layer outputs
        hidden_inputs = np.dot(self.weights_input_hidden, inputs)
        hidden_outputs = np.array([self.bipolar_step(z) for z in hidden_inputs])

        # Compute output layer
        hidden_outputs_with_bias = np.append(1, hidden_outputs)  # Add bias for output
        net_output_input = np.dot(self.weights_hidden_output, hidden_outputs_with_bias)
        final_output = self.bipolar_step(net_output_input)

        return hidden_inputs, hidden_outputs, net_output_input, final_output

    def update_weights(self, inputs, hidden_inputs, hidden_outputs, target):
        inputs = np.append(1, inputs)  # Add bias
        if target == 1:
            # Find hidden unit with net input closest to 0
            closest_hidden_index = np.argmin(np.abs(hidden_inputs))
            zinj = hidden_inputs[closest_hidden_index]
            # Update weights for the hidden unit closest to 0
            self.weights_input_hidden[closest_hidden_index] += self.learning_rate * (1 - zinj) * inputs
        elif target == -1:
            # Update weights for all hidden units with positive net input
            for j, zinj in enumerate(hidden_inputs):
                if zinj > 0:
                    self.weights_input_hidden[j] += self.learning_rate * (-1 - zinj) * inputs

        # Update weights for the hidden to output layer
        hidden_outputs_with_bias = np.append(1, hidden_outputs)
        error = target - self.bipolar_step(np.dot(self.weights_hidden_output, hidden_outputs_with_bias))
        self.weights_hidden_output += self.learning_rate * error * hidden_outputs_with_bias

    def train(self, training_data, targets, max_iterations=100):
        for iteration in range(max_iterations):
            weight_updated = False
            for sample, target in zip(training_data, targets):
                hidden_inputs, hidden_outputs, net_output_input, final_output = self.forward_propagation(sample)

                # Check if the output matches the target
                if final_output != target:
                    weight_updated = True
                    self.update_weights(sample, hidden_inputs, hidden_outputs, target)

            # Stop if no weight updates occurred in the entire epoch
            if not weight_updated:
                print(f"Training converged after {iteration + 1} iterations.")
                break
        else:
            print("Maximum iterations reached.")

    def predict(self, inputs):
        _, _, _, final_output = self.forward_propagation(inputs)
        return final_output

if __name__ == "__main__":
    # Take topology input from the user
    input_size = int(input("Enter the number of input neurons: "))
    hidden_size = int(input("Enter the number of hidden neurons: "))
    learning_rate = float(input("Enter the learning rate: "))

    print(f"Using topology: input neurons={input_size}, hidden neurons={hidden_size}, learning rate={learning_rate}")

    # Example XOR problem with bipolar inputs and outputs
    training_data = np.array([
        [1, 1],
        [1, -1],
        [-1, 1],
        [-1, -1]
    ])
    targets = np.array([-1, 1, 1, -1])  # Bipolar outputs for XOR

    madeline = MadelineNetwork(input_size=input_size, hidden_size=hidden_size, learning_rate=learning_rate)
    madeline.train(training_data, targets)

    # Test predictions
    for sample in training_data:
        prediction = madeline.predict(sample)
        print(f"Input: {sample}, Prediction: {prediction}")




# In[ ]:


# 12. GA without visualization


import random
import math

class Individual:
    def __init__(self, gnome):
        self.gnome = gnome  # The path (gnome) representing the order of cities
        self.fitness = 0  # Total distance of the path
    def __lt__(self, other):
        return self.fitness < other.fitness  # Comparison operator for fitness
    def __gt__(self, other):
        return self.fitness > other.fitness  # Comparison operator for fitness

# Helper Functions
def rand_num(start, end):
    """Generate a random number between start and end."""
    return random.randint(start, end)

def repeat(s, ch):
    """Check if a character is already in the string."""
    return ch in s

def mutated_gene(gnome, num_cities):
    """Mutate the gnome by swapping two cities randomly."""
    a, b = rand_num(1, num_cities-1), rand_num(1, num_cities-1)
    while a == b:
        b = rand_num(1, num_cities-1)
    gnome[a], gnome[b] = gnome[b], gnome[a]
    return gnome

def create_gnome(num_cities):
    """Create a valid initial gnome."""
    cities = list(range(1, num_cities))
    random.shuffle(cities)
    gnome = [0] + cities + [0]  # Start and end city are the same (0)
    return gnome

def cal_fitness(gnome, mp):
    """Calculate the total distance of a gnome."""
    distance = 0
    for i in range(len(gnome)-1):
        distance += mp[gnome[i]][gnome[i+1]]
    return distance

def cooldown(temp):
    """Reduce the temperature by a factor of 90%."""
    return temp * 0.9

def genetic_algorithm(num_cities, mp, population_size, max_generations):
    # Initialize population
    population = []
    for _ in range(population_size):
        gnome = create_gnome(num_cities)
        fitness = cal_fitness(gnome, mp)
        individual = Individual(gnome)
        individual.fitness = fitness
        population.append(individual)

    print("Initial Population:")
    for ind in population:
        print(f"Gnome: {ind.gnome}, Fitness: {ind.fitness}")

    # Initial temperature
    temp = 10000

    generation = 0
    best_solution = min(population, key=lambda x: x.fitness)

    while temp > 1000 and generation < max_generations:
        generation += 1
        # Sort Population by fitness (ascending)
        population.sort()

        # Mutation and Selection
        new_population = []
        for ind in population:
            # Mutate the gnome
            new_gnome = mutated_gene(ind.gnome[:], num_cities)
            new_fitness = cal_fitness(new_gnome, mp)
            new_individual = Individual(new_gnome)
            new_individual.fitness = new_fitness

            # Accept the new gnome if it is better or with a certain probability
            if new_fitness < ind.fitness:
                new_population.append(new_individual)
            else:
                delta_fitness = new_fitness - ind.fitness
                acceptance_probability = math.exp(-delta_fitness / temp)
                if random.random() < acceptance_probability:
                    new_population.append(new_individual)
                else:
                    new_population.append(ind)

        # Update Population
        population = new_population

        # Reduce Temperature
        temp = cooldown(temp)

        # Update the Best Solution
        best_solution = min(population, key=lambda x: x.fitness)

        # Print Current Generation
        print(f"Generation {generation}:")
        for ind in population:
            print(f"Gnome: {ind.gnome}, Fitness: {ind.fitness}")

    # Output the Best Solution
    print("\nBest Solution Found:")
    print(f"Path: {best_solution.gnome}")
    print(f"Total Distance: {best_solution.fitness}")

# Example:
num_cities = 5
mp = [
    [0, 2, 9, 10, 7],
    [1, 0, 6, 4, 3],
    [15, 7, 0, 8, 12],
    [6, 3, 12, 0, 5],
    [10, 4, 8, 3, 0]
]
population_size = 10
max_generations = 100

genetic_algorithm(num_cities, mp, population_size, max_generations)


# In[ ]:


#GA with visualization

import random
import math
import matplotlib.pyplot as plt

class Individual:
    def __init__(self, gnome):
        self.gnome = gnome  # The path (gnome) representing the order of cities
        self.fitness = 0  # Total distance of the path
    def __lt__(self, other):
        return self.fitness < other.fitness  # Comparison operator for fitness
    def __gt__(self, other):
        return self.fitness > other.fitness  # Comparison operator for fitness

# Helper Functions
def rand_num(start, end):
    """Generate a random number between start and end."""
    return random.randint(start, end)

def repeat(s, ch):
    """Check if a character is already in the string."""
    return ch in s

def mutated_gene(gnome, num_cities):
    """Mutate the gnome by swapping two cities randomly."""
    a, b = rand_num(1, num_cities-1), rand_num(1, num_cities-1)
    while a == b:
        b = rand_num(1, num_cities-1)
    gnome[a], gnome[b] = gnome[b], gnome[a]
    return gnome

def create_gnome(num_cities):
    """Create a valid initial gnome."""
    cities = list(range(1, num_cities))
    random.shuffle(cities)
    gnome = [0] + cities + [0]  # Start and end city are the same (0)
    return gnome

def cal_fitness(gnome, mp):
    """Calculate the total distance of a gnome."""
    distance = 0
    for i in range(len(gnome)-1):
        distance += mp[gnome[i]][gnome[i+1]]
    return distance

def cooldown(temp):
    """Reduce the temperature by a factor of 90%."""
    return temp * 0.9

def plot_path(cities, best_gnome):
    """Visualize the path taken by the salesman."""
    x = [cities[gnome][0] for gnome in best_gnome]
    y = [cities[gnome][1] for gnome in best_gnome]

    plt.figure(figsize=(8, 6))
    plt.plot(x, y, marker='o', color='b', linestyle='-', markersize=8, label='Path')
    plt.scatter(x[0], y[0], color='g', s=200, label="Start/End", zorder=5)
    for i in range(len(cities)):
        plt.text(cities[i][0] + 0.1, cities[i][1] + 0.1, f"City {i}")

    plt.title("Best Path Found by Genetic Algorithm (TSP)")
    plt.xlabel("X")
    plt.ylabel("Y")
    plt.legend()
    plt.grid(True)
    plt.show()

def genetic_algorithm(num_cities, mp, population_size, max_generations, cities):
    # Initialize population
    population = []
    for _ in range(population_size):
        gnome = create_gnome(num_cities)
        fitness = cal_fitness(gnome, mp)
        individual = Individual(gnome)
        individual.fitness = fitness
        population.append(individual)

    print("Initial Population:")
    for ind in population:
        print(f"Gnome: {ind.gnome}, Fitness: {ind.fitness}")

    # Initial temperature
    temp = 10000

    generation = 0
    best_solution = min(population, key=lambda x: x.fitness)

    while temp > 1000 and generation < max_generations:
        generation += 1
        # Sort Population by fitness (ascending)
        population.sort()

        # Mutation and Selection
        new_population = []
        for ind in population:
            # Mutate the gnome
            new_gnome = mutated_gene(ind.gnome[:], num_cities)
            new_fitness = cal_fitness(new_gnome, mp)
            new_individual = Individual(new_gnome)
            new_individual.fitness = new_fitness

            # Accept the new gnome if it is better or with a certain probability
            if new_fitness < ind.fitness:
                new_population.append(new_individual)
            else:
                delta_fitness = new_fitness - ind.fitness
                acceptance_probability = math.exp(-delta_fitness / temp)
                if random.random() < acceptance_probability:
                    new_population.append(new_individual)
                else:
                    new_population.append(ind)

        # Update Population
        population = new_population

        # Reduce Temperature
        temp = cooldown(temp)

        # Update the Best Solution
        best_solution = min(population, key=lambda x: x.fitness)

        # Print Current Generation
        print(f"Generation {generation}:")
        for ind in population:
            print(f"Gnome: {ind.gnome}, Fitness: {ind.fitness}")

    # Output the Best Solution
    print("\nBest Solution Found:")
    print(f"Path: {best_solution.gnome}")
    print(f"Total Distance: {best_solution.fitness}")

    # Visualize the best solution
    plot_path(cities, best_solution.gnome)

# Take input from the user
num_cities = int(input("Enter the number of cities: "))
population_size = int(input("Enter the population size: "))
max_generations = int(input("Enter the maximum number of generations: "))

# Input distance matrix
mp = []
cities = []
print("Enter the distance matrix:")
for i in range(num_cities):
    row = list(map(int, input(f"Enter distances for city {i}: ").split()))
    mp.append(row)
    x, y = map(int, input(f"Enter coordinates for city {i} (x y): ").split())
    cities.append((x, y))  # Store city coordinates for visualization

# Run the genetic algorithm
genetic_algorithm(num_cities, mp, population_size, max_generations, cities)

