"""# Importing the Libraries
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from fcmeans import FCM
from sklearn.datasets import load_iris
from sklearn.preprocessing import MinMaxScaler

# Load the Iris dataset using the pandas read_csv function
# Loading Iris dataset from sklearn as a pandas DataFrame
iris = load_iris()
data = pd.DataFrame(iris.data, columns=iris.feature_names)

#  Plot the data points on the dataset using matplotlib library
plt.figure(figsize=(8, 6))
plt.scatter(data.iloc[:, 0], data.iloc[:, 1], c='blue', label="Data points")
plt.xlabel(iris.feature_names[0])
plt.ylabel(iris.feature_names[1])
plt.title('Iris Dataset - Data Points')
plt.legend()
plt.show()

#  Initialize the fuzzy-c-means algorithm by passing the number of clusters as 2
n_clusters = 2
fcm = FCM(n_clusters=n_clusters)

#  Fit the dataset and identify the clusters
scaler = MinMaxScaler()  # Normalize the data for better clustering performance
data_scaled = scaler.fit_transform(data)
fcm.fit(data_scaled)

# Identify the cluster centers and the various labels
labels = fcm.predict(data_scaled)
centroids = fcm.centers

# Display the data points of data clusters and the various labels
data['cluster'] = labels
print("Cluster Centers: \n", centroids)
print("\nCluster Labels for Data Points:\n", data['cluster'].value_counts())

# Plot the result of the original dataset and the output for 2 clusters
plt.figure(figsize=(8, 6))
for i in range(n_clusters):
    cluster_points = data[data['cluster'] == i]
    plt.scatter(cluster_points.iloc[:, 0], cluster_points.iloc[:, 1], label=f"Cluster {i+1}")
    
# Plot centroids
for centroid in centroids:
    plt.scatter(centroid[0], centroid[1], marker='x', color='black', s=100, label="Centroid")
    
plt.xlabel(iris.feature_names[0])
plt.ylabel(iris.feature_names[1])
plt.title('Fuzzy C-Means Clustering with 2 Clusters')
plt.legend()
plt.show()

#  Iterating through the dataset for a list of clusters
cluster_range = [2, 3, 4, 5, 6, 7]
for n_clusters in cluster_range:
    fcm = FCM(n_clusters=n_clusters)
    fcm.fit(data_scaled)
    labels = fcm.predict(data_scaled)
    
    #  Find the data points for the different number of clusters
    data['cluster'] = labels
    
    #  Visualization of the Results and finding the ideal number of clusters
    plt.figure(figsize=(8, 6))
    for i in range(n_clusters):
        cluster_points = data[data['cluster'] == i]
        plt.scatter(cluster_points.iloc[:, 0], cluster_points.iloc[:, 1], label=f"Cluster {i+1}")

    plt.xlabel(iris.feature_names[0])
    plt.ylabel(iris.feature_names[1])
    plt.title(f'Fuzzy C-Means Clustering with {n_clusters} Clusters')
    plt.legend()
    plt.show()"""

    # Importing the Libraries
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from fcmeans import FCM
from sklearn.datasets import load_iris
from sklearn.preprocessing import MinMaxScaler

#  Load the Iris dataset using the pandas read_csv function
# Loading Iris dataset from sklearn as a pandas DataFrame
iris = load_iris()
data = pd.DataFrame(iris.data, columns=iris.feature_names)

#  Plot the data points on the dataset using matplotlib library
plt.figure(figsize=(8, 6))
plt.scatter(data.iloc[:, 0], data.iloc[:, 1], c='blue', label="Data points")
plt.xlabel(iris.feature_names[0])
plt.ylabel(iris.feature_names[1])
plt.title('Iris Dataset - Data Points')
plt.legend()
plt.show()

# Initialize the fuzzy-c-means algorithm by passing the number of clusters as 3
n_clusters = 3  # Set the number of clusters to 3
fcm = FCM(n_clusters=n_clusters)

# Fit the dataset and identify the clusters
scaler = MinMaxScaler()  # Normalize the data for better clustering performance
data_scaled = scaler.fit_transform(data)
fcm.fit(data_scaled)

#  Identify the cluster centers and the various labels
labels = fcm.predict(data_scaled)
centroids = fcm.centers

#Display the data points of data clusters and the various labels
data['cluster'] = labels
print("Cluster Centers: \n", centroids)
print("\nCluster Labels for Data Points:\n", data['cluster'].value_counts())

# Plot the result of the original dataset and the output for 3 clusters
plt.figure(figsize=(8, 6))
for i in range(n_clusters):
    cluster_points = data[data['cluster'] == i]
    plt.scatter(cluster_points.iloc[:, 0], cluster_points.iloc[:, 1], label=f"Cluster {i+1}")

# Plot centroids
for centroid in centroids:
    plt.scatter(centroid[0], centroid[1], marker='x', color='black', s=100, label="Centroid")

plt.xlabel(iris.feature_names[0])
plt.ylabel(iris.feature_names[1])
plt.title(f'Fuzzy C-Means Clustering with {n_clusters} Clusters')
plt.legend()
plt.show()


